//File: TagFix.cpp
//Author: Colin Kern (colin.kern AT gmail DOT edu)

#include "TagFix.h"
#include "ScriptProcess.h"
#include <QFileDialog>
#include <QProcess>
#include <QStringList>
#include <QTreeView>
#include <QStandardItemModel>
#include <QMessageBox>
#include <QLabel>
#include <QProgressBar>
#include <QPushButton>
#include <QSettings>
#include <QDirIterator>
#include <fileref.h>
#include <tag.h>

TagFix::TagFix(QWidget *parent)
    : QMainWindow(parent)
{
	ui.setupUi(this);
	
	model = new QStandardItemModel(0, 5);
	model->setHorizontalHeaderLabels(QStringList() << "Confidence" << "New Artist" << "New Title" << "Original Artist" << "Original Track");
	
	treeView = new QTreeView();
	treeView->setModel(model);
	treeView->setAnimated(true);
	treeView->setSortingEnabled(true);
	
	queryCount = 0;
	
	createStatusBar();
	
	QSettings settings("TagFix","TagFix");
	scriptPath = settings.value("scriptPath", QString()).toString();
	
	connect(ui.actionAbout_Qt, SIGNAL(triggered()), this, SLOT(aboutQt()));
	connect(ui.actionProcess_Files, SIGNAL(triggered()), this, SLOT(chooseFile()));
	connect(ui.actionProcess_Directory, SIGNAL(triggered()), this, SLOT(chooseDirectory()));
	connect(ui.action_Quit, SIGNAL(triggered()), this, SLOT(close()));
	connect(ui.actionSet_Script_Path, SIGNAL(triggered()), this, SLOT(setScriptPath()));
	
	setCentralWidget(treeView);
	updateTitleAndStatus();
}

TagFix::~TagFix()
{

}

void TagFix::aboutQt()
{
	QApplication::aboutQt();
}

void TagFix::chooseFile()
{
	QStringList filenames = QFileDialog::getOpenFileNames(this, "Choose MP3s", QString(), tr("Music Files (*.mp3)"));
	QStringListIterator i(filenames);
	while (i.hasNext()) {
		queue.append(new QPair<QString, int>(i.next(), 0));
	}
	failedFiles.clear();
	progress->setMaximum(queue.size());
	sendNextQuery();
}

void TagFix::chooseDirectory()
{
	QString dirName = QFileDialog::getExistingDirectory(this, "Choose Directory");
	if (!dirName.isEmpty()) {
		QDir dir(dirName);
		addDirectoryToQueue(dir);
		failedFiles.clear();
		progress->setMaximum(queue.size());
		sendNextQuery();
	}
}

void TagFix::addDirectoryToQueue(QDir &dir)
{
	//Add files
	dir.setFilter(QDir::Files);
	dir.setNameFilters(QStringList() << "*.mp3");
	QDirIterator i(dir, QDirIterator::Subdirectories);
	while (i.hasNext()) {
		queue.append(new QPair<QString, int>(i.next(), 0));
	}
	
}

void TagFix::scanComplete(ScriptProcess *process, QProcess::ExitStatus status) 
{
	if (status == QProcess::NormalExit) {
		QString data = QString(process->readAllStandardOutput());
		if (data.isEmpty()) {
			QMessageBox::critical(this, "Error", "Script returned errors:\n" + process->readAllStandardError());
		} else {
			bool different = false;
			QStringList tracks = data.split("<track");
			tracks.removeFirst();
			QStringListIterator i(tracks);
			if (i.hasNext()) {
				QList<QStandardItem*> items = extractDataItem(i.next());
				TagLib::FileRef ref(QFile::encodeName(process->getFilePath()).data());
				TagLib::String artist = ref.tag()->artist();
				items << new QStandardItem(TStringToQString(artist));
				if (items.last()->text() == items[1]->text()) {
					items.last()->setForeground(QBrush(Qt::green));
				} else {
					items.last()->setForeground(QBrush(Qt::red));
					items[0]->setCheckState(Qt::Checked);
					different = true;
				}
				TagLib::String title = ref.tag()->title();
				items << new QStandardItem(TStringToQString(title));
				if (items.last()->text() == items[2]->text()) {
					items.last()->setForeground(QBrush(Qt::green));
				} else {
					items.last()->setForeground(QBrush(Qt::red));
					items[0]->setCheckState(Qt::Checked);
					different = true;
				}
				items << new QStandardItem(process->getFilePath());
				if (different) {
					model->appendRow(items);
					while (i.hasNext()) {
						items[0]->appendRow(extractDataItem(i.next()));
					}
				}
			} else {
				if (process->attempts() > 4) {
					failedFiles.append(process->getFilePath());
				} else {
					queue.prepend(new QPair<QString, int>(process->getFilePath(), process->attempts() + 1));
				}
			}
		}
	} else {
		QMessageBox::critical(this, "Crashed", "The process running the query crashed.");
	}
	queryCount--;
	sendNextQuery();
}

void TagFix::error(QProcess::ProcessError error)
{
	if (error == QProcess::FailedToStart) {
		QMessageBox::critical(this, "Error", "Script failed to start");
	} else {
		QMessageBox::critical(this, "Error", "Script runtime error");
	}
}

void TagFix::setScriptPath()
{
	QString newPath = QFileDialog::getOpenFileName(this, "Select Script Location");
	if (!newPath.isEmpty()) {
		QSettings settings("TagFix", "TagFix");
		settings.setValue("scriptPath", newPath);
		scriptPath = newPath;
	}
}

void TagFix::stopJobs()
{
	queue.clear();
}

void TagFix::saveTags()
{
	QString string;
	int counter = 0;
	for (int i = 0; i < model->rowCount(); ++i) {
		if (model->item(i, 0)->checkState() == Qt::Checked) {
			counter++;
		}
	}
	int result = QMessageBox::question(this, "TagFix",
			"Are you sure you want to edit the tags of " + QString::number(counter) + " tracks?",
			   QMessageBox::Yes,
			   QMessageBox::Cancel | QMessageBox::Default | QMessageBox::Escape);
	if (result == QMessageBox::Cancel) {
		return;
	}
	for (int i = 0; i < model->rowCount(); ++i) {
		if (model->item(i, 0)->checkState() == Qt::Checked) {
			TagLib::FileRef ref(model->item(i, 5)->text().toUtf8().data());
			ref.tag()->setArtist(model->item(i, 1)->text().toUtf8().data());
			ref.tag()->setTitle(model->item(i, 2)->text().toUtf8().data());
			ref.save();
		}
	}
	QMessageBox::information(this, "TagFix", "Done updating " + QString::number(counter) + " tracks");
}

QList<QStandardItem*> TagFix::extractDataItem(const QString &track) const
{
	QList<QStandardItem*> data;
	QRegExp re = QRegExp("confidence\\=\"(0\\.[0-9]*)\"");
	if (re.indexIn(track) > -1) {
		QStandardItem *confidence = new QStandardItem(re.cap(1));
		confidence->setCheckable(true);
		confidence->setTristate(true);
		data << confidence;
	}
	re = QRegExp("<artist>(.*)</artist>");
	if (re.indexIn(track) > -1) {
		data << new QStandardItem(re.cap(1));
	}
	re = QRegExp("<title>(.*)</title>");
	if (re.indexIn(track) > -1) {
		data << new QStandardItem(re.cap(1));
	}
	return data;
}

void TagFix::scanFile(QPair<QString, int> *filename)
{
	ScriptProcess *script = new ScriptProcess(filename->first, filename->second, this);
	connect(script, SIGNAL(done(ScriptProcess*, QProcess::ExitStatus)), this, SLOT(scanComplete(ScriptProcess*, QProcess::ExitStatus)));
	connect(script, SIGNAL(error(QProcess::ProcessError)), this, SLOT(error(QProcess::ProcessError)));
	QStringList arguments;
	arguments << filename->first;
	script->start(scriptPath, arguments);
	queryCount++;
}

void TagFix::createStatusBar()
{
	statusLabel = new QLabel();
	progress = new QProgressBar();
	progress->setTextVisible(false);
	stopButton = new QPushButton("Stop");
	saveButton = new QPushButton("Save");
	
	connect(stopButton, SIGNAL(clicked()), this, SLOT(stopJobs()));
	connect(saveButton, SIGNAL(clicked()), this, SLOT(saveTags()));
	
	statusBar()->addWidget(progress);
	statusBar()->addWidget(statusLabel);
	statusBar()->addWidget(stopButton);
	statusBar()->addWidget(saveButton);
}

void TagFix::sendNextQuery()
{
	if (queryCount < 5 && !queue.isEmpty()) {
		scanFile(queue.front());
		queue.removeFirst();
		sendNextQuery();
	}
	updateTitleAndStatus();
}

void TagFix::updateTitleAndStatus()
{
	statusLabel->setText("Running " + QString::number(queryCount) + " queries, " + QString::number(queue.size()) + " in queue");
	setWindowTitle("TagFix - " + QString::number(queue.size()) + " in queue");
	progress->setValue(queue.size() + queryCount);
	if (queue.empty() && queryCount == 0) {
		stopButton->hide();
		saveButton->show();
		if (failedFiles.size() > 0) {
			QMessageBox::critical(this, "Error", 
				"Could not get data for " + QString::number(failedFiles.size()) + " tracks"); 
		}
	} else {
		saveButton->hide();
		stopButton->show();
	}
}

